package kiev.stdlib;

import kiev.stdlib.Pair.*;
import kiev.stdlib.NoSuchElementException;
import kiev.stdlib.Enumeration;

/** a class for immutable lists
 * @see		ListBuffer
 * @author      Martin Odersky
 */

public class List<A> {

    public case Nil;
    public case Cons/*<A>*/(A head, List<A> tail);

/** return the length of the list
 */
    public int length() {
	switch (this) {
	case Nil:
	    return 0;
	case Cons(_, List<A> xs):
	    return 1 + xs.length();
	}
    }

/** the first element of the list
 */
    public A head() {
	switch (this) {
	case Nil:
	    throw new MatchError("Nil.head");
	case Cons(A x, List<A> xs):
	    return x;
	}
    }

/** the rest without the first element of the list
 */
    public List<A> tail() {
	switch (this) {
	case Nil:
	    throw new MatchError("Nil.tail");
	case Cons(A x, List<A> xs):
	    return xs;
	}
    }

/** the list consisting of the first `n' elements of this list
 */
    public List<A> take(int n) {
	if (n == 0)
	    return Nil;
	else
	    switch (this) {
	    case Nil:
		return this;
	    case Cons(A x, List<A> xs):
		return new Cons<A,List<A>>(x, xs.take(n-1));
	    }
    }

/** the list without its first `n' elements
 */
    public List<A> drop(int n) {
	if (n == 0)
	    return this;
	else
	    switch (this) {
	    case Nil:
		return this;
	    case Cons(A x, List<A> xs):
		return xs.drop(n-1);
	    }
    }

/** split lists into two. The first part of the results is a list
 *  which consists of the first `n' elements of this list. The second part 
 *  is a list containing all remaining elements.
 *  @law    split(n, xs) = Pair.Pair(take(n, xs), drop(n, xs))
 */
//    public Pair<List<A>, List<A>> split(int n) {
//	if (n == 0)
//	    return new Pair(Nil, this);
//	else
//	    switch (this) {
//	    case Nil:
//		return new Pair(Nil, Nil);
//	    case Cons(A x, List<A> xs):
//		Pair<List<A>, List<A>> p = xs.split(n-1);
//	        p.fst = new Cons<A,List<A>>(x, p.fst);
//		return p;
//	    }
//   }

/** the element at position `n' in this list
 */
    public A at(int n) {
	switch (this) {
	case Nil:
	    throw new MatchError("Nil.at");
	case Cons(A x, List<A> xs):
	    if (n == 0) return x;
	    else return xs.at(n-1);
	}
    }

/** the last element of this list
 */
    public A last() {
	return at(length() - 1);
    }

/** all elements except the last one
 */
    public List<A> init() {
	return take(length() - 1);
    }
  
/** the elements in reverse order
 */
    public List<A> reverse() {
	return reverse(this);
    }

    private List<A> reverse(List<A> xs) {
	List<A> ys = Nil;
	while (true) {
	    switch (xs) {
	    case Nil:
		return ys;
	    case Cons(A x, List<A> xs1):
		ys = new Cons<A,List<A>>(x, ys);
	        xs = xs1;
		break;
	    }
	}
    }

/** return the result of appending element `y' to this list as a new list
 */
    public List<A> concat(A y) {
	return concat(new Cons<A,List<A>>(y, Nil));
    }

/** return the result of appending list `ys' to this list as a new list
 */
    public List<A> concat(List<A> ys) {
	switch (this) {
	case Nil:
	    return ys;
	case Cons(A x, List<A> xs):
	    return new Cons<A,List<A>>(x, xs.concat(ys));
	}
    }

/** concatenate all element lists of `xss'
 */ 
//    public static <A> List<A> concatenate(List<List<A>> xss) {
//	switch (xss) {
//	case Nil:
//	    return Nil;
//	case Cons(List<A> xs, List<List<A>> xss1):
//	    return xs.concat(concatenate(xss1));
//	}
//    }

/** apply `f' to each element of this list
 */
//    public <B> void forall((A)->B f) {
//	List<A> xs = this;
//	while (true) {
//	    switch (xs) {
//	    case Nil:
//		return;
//	    case Cons(A x, List<A> xs1):
//		f(x);
//	        xs = xs1;
//	    }
//	}
//    }

/** apply `f' to each element of this list and collect the results in a
 *  new list 
 */
//    public <B> List<B> map((A)->B f) {
//	switch (this) {
//	case Nil:
//	    return Nil;
//	case Cons(A x, List<A> xs):
//	    return new Cons<A,List<A>>(f(x), xs.map(f));
//	}
//    }

/** apply `f' to each element of this list and concatenate the (list-valued)
 *  results in a new list 
 */
//    public <B> List<B> bind((A)->List<B> f) {
//	switch (this) {
//	case Nil:
//	    return Nil;
//	case Cons(A x, List<A> xs):
//	    return f(x).concat(xs.bind(f));
//	}
//    }
    
/** all elements for which predicate`p' is true
 */
    public List<A> filter((A)->boolean p) {
	switch (this) {
	case Nil:
	    return this;
	case Cons(A x, List<A> xs):
	    if (p(x)) return new Cons<A,List<A>>(x, xs.filter(p));
	    else return xs.filter(p);
	}
    }

/** fold left:
 *
 *  @law List.cons(x1,...,xN).foldl(OP, z) = (...(z OP x1) OP ... ) OP xN
 *
 */
//    public <B> B foldl((B,A)->B f, B z) {
//	List<A> xs = this;
//	while (true) {
//	    switch (xs) {
//	    case Nil:
//		return z;
//	    case Cons(A x, List<A> xs1):
//		z = f(z, x);
//	        xs = xs1;
//	    }
//	}
//    }

/** fold right:
 *
 *  @law List.cons(x1,...,xN).foldr(OP, z) = x1 OP (... OP (xN OP z)...)
 *
 */
//    public <B> B foldr((A,B)->B f, B z) {
//	switch (this) {
//	case Nil:
//	    return z;
//	case Cons(A x, List<A> xs):
//	    return f(x, xs.foldr(f, z));
//	}
//    }

/** does this list contain element `y'?
 */
    public boolean contains(A y) {
	switch (this) {
	case Nil:
	    return false;
	case Cons(A x, List<A> xs):
	    return ((Object)x).equals((Object)y) || xs.contains(y);
	}
    }

/** this list without the elements in `ys'. Each element is removed
 *  at most once.
 */
    public List<A> diff(List<A> ys) {
	switch (ys) {
	case Nil:
	    return this;
	case Cons(A y, List<A> ys1):
	    return this.diff(y).diff(ys1);
	}
    }

/** this list without the element`y'. The element `y' is removed
 *  at most once.
 */
    public List<A> diff(A y) {
	switch (this) {
	case Nil:
	    return this;
	case Cons(A x, List<A> xs):
	    if (((Object)x).equals((Object)y)) {
		return xs;
	    } else {
		List<A> xs1 = xs.diff(y);
		if (xs1 == xs) return this;
		else return new Cons<A,List<A>>(x, xs1);
	    }
	}
    }

    public int hashCode() {
	switch (this) {
	case Nil:
	    return 0;
	case Cons(A x, List<A> xs):
	    return ((Object)x).hashCode() * 41 + ((Object)xs).hashCode();
	}
    }

    public boolean equals(Object other) {
	if (other instanceof List) {
	    switch (this) {
	    case Nil:
		return (List)other == Nil;
	    case Cons(A x, List<A> xs):
		switch ((List)other) {
		case Nil:
		    return false;
//		case <B> Cons(B y, List<B> ys):
// TODO - A should be B!
//		case Cons(A y, List<A> ys):
		case Cons(_,_): {
		    Object y = (Object)((Cons)other).head;
		    Object ys = (Object)((Cons)other).tail;
		    return ((Object)x).equals((Object)y) && 
			   ((Object)xs).equals((Object)ys);
		}
		}
	    }
	} else {
	    return false;
	}
    }

/** a String representation of this list
 */
    public String toString() {
	return "List(" + elementsToString(", ") + ")";
    }

/** a String representation of the elements in this list with `sep' used
 *  as a separator
 */
    public String elementsToString(String sep) {
	switch (this) {
	case Nil:
	    return "";
	case Cons(A x, List<A> xs):
	    return ((Object)x).toString() + xs.restToString(sep);
	}
    }

    private String restToString(String sep) {
	switch (this) {
	case Nil:
	    return "";
	case Cons(A x, List<A> xs):
	    return sep + ((Object)x).toString() + xs.restToString(sep);
	}
    }

/** return the elements of this list as an enumeration
 */
    public Enumeration<A> elements() {
	return new ListEnumerator/*<A>*/(this);
    }

/** zip two lists into a list of pairs
 */
//    public List<Pair<A,A>> zip(List<A> ys) {
//	switch (Pair.Pair(this, ys)) {
//	case Pair(Nil, _):
//	case Pair(_, Nil):
//	    return Nil;
//	case Pair(Cons(A x, List<A> xs1), Cons(A y, List<A> ys1)):
//	    return new Cons<A,List<A>>(Pair.Pair(x, y), xs1.zip(ys1));
//	}
//    }

/** the elements of array `elems' as a list
 */
//    public static <A> List<A> fromArray(A[] elems) {
//	List<A> l = Nil;
//	int i = elems.length;
//	while (i > 0) {
//	    i--;
//	    l = new Cons<A,List<A>>(elems[i], l);
//	}
//	return l;
//    }

/** copy this list into array `elems'. return `elems' as a result.
 */
    public A[] copy(A[] elems) {
	return copy(elems, 0);
    }

/** copy this list into array `elems' staring from given `offset'.
 *  return `elems' as a result.
 */
    public A[] copy(A[] elems, int offset) {
	switch (this) {
	case Nil:
	    return elems;
	case Cons(A x, List<A> xs):
	    elems[offset] = x;
	    return xs.copy(elems, offset + 1);
	}
    }

/** list's consisting of given (0-10) elements
 */
//	static public <A> List<A> cons() {
//		return Nil;
//	}

//    static public <A> List<A> cons(A x1) {
//	return new Cons<A,List<A>>(x1, Nil);
//    }

//    static public <A> List<A> cons(A x1, A x2) {
//	return new Cons<A,List<A>>(x1, new Cons<A,List<A>>(x2, Nil));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3) {
//	return Cons(x1, Cons(x2, Cons(x3, Nil)));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Nil))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, Nil)))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5,
//				   A x6) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, 
//	       Cons(x6, Nil))))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5,
//				   A x6, A x7) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, 
//	       Cons(x6, Cons(x7, Nil)))))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5,
//				   A x6, A x7, A x8) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, 
//	       Cons(x6, Cons(x7, Cons(x8, Nil))))))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5,
//				   A x6, A x7, A x8, A x9) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, 
//	       Cons(x6, Cons(x7, Cons(x8, Cons(x9, Nil)))))))));
//    }

//    static public <A> List<A> cons(A x1, A x2, A x3, A x4, A x5,
//				   A x6, A x7, A x8, A x9, A x10) {
//	return Cons(x1, Cons(x2, Cons(x3, Cons(x4, Cons(x5, 
//	       Cons(x6, Cons(x7, Cons(x8, Cons(x9, Cons(x10, Nil))))))))));
//    }

	


}

class ListEnumerator<A> implements Enumeration<A>, Cloneable {

	private List<A> current;

	public ListEnumerator(List<A> xs) {
	    current = xs;
	}

	public boolean hasMoreElements() {
	    return current != List.Nil;
	}

	public A nextElement() {
	    switch (current) {
	    case List.Cons(A x, List<A> xs):
		current = xs;
		return x;
	    case List.Nil:
		throw new NoSuchElementException();
	   }
	}
}

    

